import array
import errno
import socket

class IO(object):
  """ Base class for handling socket communication to the Kafka server. """

  def __init__(self, host='localhost', port=9092):
    self.socket = None

    #: Hostname to connect to.
    self.host   = host

    #: Port to connect to.
    self.port   = port

  def connect(self):
    """ Connect to the Kafka server. """

    self.socket = socket.socket()
    self.socket.connect((self.host, self.port))

  def reconnect(self):
    """ Reconnect to the Kafka server. """
    self.disconnect()
    self.connect()

  def disconnect(self):
    """ Disconnect from the remote server & close the socket. """
    try:
      self.socket.close()
    except IOError:
      pass
    finally:
      self.socket = None

  def read(self, length):
    """ Send a read request to the remote Kafka server. """

    # Create a character array to act as the buffer.
    buf         = bytearray(length)
    bytes_left  = length

    try:
      while bytes_left > 0:
        read_length = self.socket.recv_into(memoryview(buf)[length-bytes_left:], bytes_left)
        bytes_left -= read_length

    except errno.EAGAIN:
      self.disconnect()
      raise IOError, "Timeout reading from the socket."

    else:
      return str(buf)

  def write(self, data):
    """ Write `data` to the remote Kafka server. """

    if self.socket is None:
      self.reconnect()

    wrote_length = 0

    try:
      wrote_length = self.__write(data)

    except (errno.ECONNRESET, errno.EPIPE, errno.ECONNABORTED):
      # Retry once.
      self.reconnect()
      wrote_length = self.__write(data)

    finally:
      return wrote_length

  def __write(self, data):
    write_length = len(data)
    wrote_length = 0

    while write_length > wrote_length:
      wrote_length += self.socket.send(data)

    return wrote_length



import atexit
import contextlib
import itertools
import struct
import threading

import kafka.io
import kafka.request_type


class Producer(kafka.io.IO):
  """Class for sending data to a `Kafka <http://sna-projects.com/kafka/>`_ broker.

    :param topic: The topic to produce to.
    :param partition: The broker partition to produce to.
    :param host: The kafka host.
    :param port: The kafka port.
    :param max_message_sz: The maximum allowed size of a produce request (in bytes). [default: 1MB]

  """

  PRODUCE_REQUEST_ID = kafka.request_type.PRODUCE

  def __init__(self, topic, partition=0, host='localhost', port=9092, max_message_sz=1048576):
    kafka.io.IO.__init__(self, host, port)
    self.max_message_sz = max_message_sz
    self.topic = topic
    self.partition = partition
    self.connect()

  def _pack_payload(self, messages):
    """Pack a list of messages into a sendable buffer.

      :param msgs: The packed messages to send.
      :param size: The size (in bytes) of all the `messages` to send.

    """
    payload = ''.join(messages)
    payload_sz = len(payload)
    topic_sz = len(self.topic)
    # Create the request as::
    #   <REQUEST_ID: short>
    #   <TOPIC_SIZE: short>
    #   <TOPIC: bytes>
    #   <PARTITION: int>
    #   <BUFFER_SIZE: int>
    #   <BUFFER: bytes>
    return struct.pack(
      '>HH%dsii%ds' % (topic_sz, payload_sz),
      self.PRODUCE_REQUEST_ID,
      topic_sz,
      self.topic,
      self.partition,
      payload_sz,
      payload
    )

  def _pack_kafka_message(self, payload):
    """Pack a payload in a format kafka expects."""
    return struct.pack('>i%ds' % len(payload), len(payload), payload)

  def encode_request(self, messages):
    """Encode a sequence of messages for sending to a kafka broker.

      Encoding a request can yeild multiple kafka messages if the payloads exceed
      the maximum produce size.

      :param messages: An iterable of :class:`Message <kafka.message>` objecjts.
      :rtype: A generator of packed kafka messages ready for sending.

    """
    encoded_msgs = []
    encoded_msgs_sz = 0
    for message in messages:
      encoded = message.encode()
      encoded_sz = len(encoded)
      if encoded_sz + encoded_msgs_sz > self.max_message_sz:
        yield self._pack_kafka_message(self._pack_payload(encoded_msgs))
        encoded_msgs = []
        encoded_msgs_sz = 0
      msg = struct.pack('>i%ds' % encoded_sz, encoded_sz, encoded)
      encoded_msgs.append(msg)
      encoded_msgs_sz += encoded_sz
    if encoded_msgs:
      yield self._pack_kafka_message(self._pack_payload(encoded_msgs))

  def send(self, messages):
    """Send a :class:`Message <kafka.message>` or a sequence of `Messages` to the Kafka server."""
    if isinstance(messages, kafka.message.Message):
      messages = [messages]
    for message in self.encode_request(messages):
      sent = self.write(message)
      if sent != len(message):
        raise IOError('Failed to send kafka message - sent %s/%s many bytes.' % (sent, len(message)))

  @contextlib.contextmanager
  def batch(self):
    """Send messages with an implict `send`."""
    messages = []
    yield(messages)
    self.send(messages)


class BatchProducer(Producer):
  """Class for batching messages to a `Kafka <http://sna-projects.com/kafka/>`_ broker with periodic flushing.

    :param topic: The topic to produce to.
    :param batch_interval: The amount of time (in seconds) to wait for messages before sending.
    :param partition: The broker partition to produce to.
    :param host: The kafka host.
    :param port: The kafka port.

  """

  PRODUCE_REQUEST_ID = kafka.request_type.PRODUCE

  def __init__(self, topic, batch_interval, partition=0, host='localhost', port=9092):
    Producer.__init__(self, topic, partition=partition, host=host, port=port)
    self.batch_interval = batch_interval
    self._message_queue = []
    self.event = threading.Event()
    self.lock = threading.Lock()
    self.timer = threading.Thread(target=self._interval_timer)
    self.timer.daemon = True
    self.timer.start()
    self.connect()
    atexit.register(self.close)

  def _interval_timer(self):
    """Flush the message queue every `batch_interval` seconds."""
    while not self.event.is_set():
      self.flush()
      self.event.wait(self.batch_interval)

  def enqueue(self, message):
    """Enqueue a message in the queue.

      .. note:: These messages are implicitly sent every `batch_interval` seconds.

      :param message: The message to queue for sending at next send interval.

    """
    with self.lock:
      self._message_queue.append(message)

  def flush(self):
    """Send all messages in the queue now."""
    with self.lock:
      if len(self._message_queue) > 0:
        self.send(self._message_queue)
        # Reset the queue
        del self._message_queue[:]

  def close(self):
    """Shutdown the timer thread and flush the message queue."""
    self.event.set()
    self.flush()
    self.timer.join()



#!/usr/bin/env python

"""
# pykafka

pykafka allows you to produce messages to the Kafka distributed publish/subscribe messaging service.

## Requirements

You need to have access to your Kafka instance and be able to connect through
TCP. You can obtain a copy and instructions on how to setup kafka at
https://github.com/kafka-dev/kafka

## Installation
pip install pykafka

## Usage

### Sending a simple message

    import kafka

    producer = kafka.producer.Producer('test')
    message  = kafka.message.Message("Foo!")
    producer.send(message)

### Sending a sequence of messages

    import kafka

    producer = kafka.producer.Producer('test')
    message1 = kafka.message.Message("Foo!")
    message2 = kafka.message.Message("Bar!")
    producer.send([message1, message2])

### Batching a bunch of messages using a context manager.

    import kafka
    producer = kafka.producer.Producer('test')

    with producer.batch() as messages:
      print "Batching a send of multiple messages.."
      messages.append(kafka.message.Message("first message to send")
      messages.append(kafka.message.Message("second message to send")

* they will be sent all at once, after the context manager execution.

### Consuming messages one by one

    import kafka
    consumer = kafka.consumer.Consumer('test')
    messages = consumer.consume()

### Consuming messages using a generator loop

    import kafka

    consumer = kafka.consumer.Consumer('test')

    for message in consumer.loop():
      print message

Contact:

Please use the GitHub issues: https://github.com/dsully/pykafka/issues

* Inspiried from Alejandro Crosa's kafka-rb: https://github.com/acrosa/kafka-rb
"""

import setuptools

# Don't install deps for development mode.
setuptools.bootstrap_install_from = None

setuptools.setup(
  name = 'pykafka',
  version = '0.1.3',
  license = 'MIT',
  long_description = __doc__,
  author = "Dan Sully",
  author_email = "dsully@gmail.com",
  url = 'http://github.com/dsully/pykafka',
  platforms = 'any',

  # What are we packaging up?
  packages = setuptools.find_packages('.'),

  zip_safe = True,
  verbose = False,
)



import struct
import time

import kafka.io
import kafka.request_type

class Consumer(kafka.io.IO):

  CONSUME_REQUEST_TYPE = kafka.request_type.FETCH

  MAX_SIZE = 1024 * 1024

  # seconds.
  DEFAULT_POLLING_INTERVAL = 2

  def __init__(self, topic, partition=0, host='localhost', port=9092):
    kafka.io.IO.__init__(self, host, port)

    #: The topic queue to consume.
    self.topic        = topic

    #: The partition the topic queue is on.
    self.partition    = partition

    #: Offset in the Kafka queue in bytes?
    self.offset       = 0

    #: Maximum message size to consume.
    self.max_size     = self.MAX_SIZE
    self.request_type = self.CONSUME_REQUEST_TYPE
    self.polling      = self.DEFAULT_POLLING_INTERVAL

    self.connect()

  def consume(self):
    """ Consume data from the topic queue. """

    self.send_consume_request()

    return self.parse_message_set_from(self.read_data_response())

  def loop(self):
    """ Loop over incoming messages from the queue in a blocking fashion. Set `polling` for the check interval in seconds. """

    while True:
      messages = self.consume()

      if messages and isinstance(messages, list) and len(messages) > 0:
        for message in messages:
          yield message

      time.sleep(self.polling)

  # REQUEST TYPE ID + TOPIC LENGTH + TOPIC + PARTITION + OFFSET + MAX SIZE
  def request_size(self):
    return 2 + 2 + len(self.topic) + 4 + 8 + 4

  def encode_request_size(self):
    return struct.pack('>i', self.request_size())

  def encode_request(self):
    length = len(self.topic)

    return struct.pack('>HH%dsiQi' % length, self.request_type, length, self.topic, self.partition, self.offset, self.max_size)

  def send_consume_request(self):
    self.write(self.encode_request_size())
    self.write(self.encode_request())

  def read_data_response(self):
    buf_length = struct.unpack('>i', self.read(4))[0]

    # Start with a 2 byte offset
    return self.read(buf_length)[2:]

  def parse_message_set_from(self, data):
    messages  = []
    processed = 0
    length    = len(data) - 4

    while (processed <= length):
      message_size = struct.unpack('>i', data[processed:processed+4])[0]
      messages.append(kafka.message.parse_from(data[processed:processed + message_size + 4]))
      processed += 4 + message_size

    self.offset += processed

    return messages



import struct
import zlib

def parse_from(binary):
  """ Turn a packed binary message as recieved from a Kafka broker into a :class:`Message`. """

  # A message. The format of an N byte message is the following:
  # 1 byte "magic" identifier to allow format changes
  # 4 byte CRC32 of the payload
  # N - 5 byte payload
  size     = struct.unpack('>i', binary[0:4])[0]
  magic    = struct.unpack('>B', binary[4:5])[0]
  checksum = struct.unpack('>i', binary[5:9])[0]
  payload  = binary[9:9+size]

  return Message(payload, magic, checksum)

class Message(object):
  """ A Kafka Message object. """

  MAGIC_IDENTIFIER_DEFAULT = 0

  def __init__(self, payload=None, magic=MAGIC_IDENTIFIER_DEFAULT, checksum=None):
    self.magic    = magic
    self.checksum = checksum
    self.payload  = None

    if payload is not None:
      self.payload = str(payload)

    if self.payload is not None and self.checksum is None:
      self.checksum = self.calculate_checksum()

  def __str__(self):
    return self.payload

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.magic == other.magic and self.payload == other.payload and self.checksum == other.checksum

    return False

  def __ne__(self, other):
    return not self.__eq__(other)

  def calculate_checksum(self):
    """ Returns the checksum for the payload. """

    return zlib.crc32(self.payload)

  def is_valid(self):
    """ Returns true if the checksum for this message is valid. """
    return self.checksum == self.calculate_checksum()

  def encode(self):
    """ Encode a :class:`Message` to binary form. """

    # <MAGIC_BYTE: char> <CRC32: int> <PAYLOAD: bytes>
    return struct.pack('>Bi%ds' % len(self.payload), self.magic, self.calculate_checksum(), self.payload)

